import EventEmitter from 'events';
import * as fs from 'fs';
import * as path from 'path';
import { sequence } from './utils/promise.js';
import { assign } from './utils/object.js';
import { name, version } from '../package.json';
import checkVersion from './utils/checkVersion.js';

const opts = { encoding: 'utf-8', persistent: true };
const STYLE_ECXT_REG = /\.(less|sass|scss|css)$/;
 
class FileWatcher {
  constructor(file, data, callback, dispose) {
    try {
      const fsWatcher = fs.watch(file, opts, event => {
        if (event === 'rename') {
          fsWatcher.close();
          dispose();
          callback(file);
        } else {
          // this is necessary because we get duplicate events...
          const contents = fs.readFileSync(file, 'utf-8');
          if (contents !== data) {
            data = contents;
            callback(file);
          }
        }
      });

      this.fileExists = true;
    } catch (err) {
      if (err.code === 'ENOENT') {
        // can't watch files that don't exist (e.g. injected
        // by plugins somehow)
        this.fileExists = false;
      } else {
        throw err;
      }
    }
  }
}

export default function watch(rollup, options) {
  const emitter = new EventEmitter();

  process.nextTick(() => emitter.emit('event', { code: 'STARTING' }));

  checkVersion(name, version)
    .catch(err => {
      if (err.code === 'OUT_OF_DATE') {
        // TODO offer to update
        console.error(`rollup-watch is out of date (you have ${err.localVersion}, latest version is ${err.latestVersion}). Update it with npm install -g rollup-watch`); // eslint-disable-line no-console
      }
    })
    .then(() => {
      let filewatchers = new Map();

      let rebuildScheduled = false;
      let building = false;
      let watching = false;

      let timeout;
      let cache;

      var htmlCSSCache = new Map();

      function triggerRebuild(id) {
        clearTimeout(timeout);
        rebuildScheduled = true;

        console.log(`[File Changed]: ${id}`);

        timeout = setTimeout(function () {
          if (!building) {
            rebuildScheduled = false;

            // clear cache refrenced the changing html/css files
            if (htmlCSSCache.has(id)) {
              var relatedIds = htmlCSSCache.get(id).relatedIds;
              cache.modules = cache.modules.filter(function (module) {
                return !relatedIds.has(module.id);
              });
            }

            build();
          }
        }, 100);
      }
      
      function extractImportsFromCSS(module) {
        var id = module.id;
        var dir = path.dirname(id);
        var source = module.originalCode;

        var temp = id.match(STYLE_ECXT_REG);
        var extension = temp && temp[0];

        if (!extension) {
          throw `[style file extension invalid]: ${id}`;
        }

        // css/scss/less
        // not very strict
        var importRegex = /\s*@?import\s+(['"])([^'"]+)\1/g;
        var match = null;
        var pathArray = [];

        while (match = importRegex.exec(source)) {
          var url = match[2];
          var exists = false;

          if (!STYLE_ECXT_REG.test(url)) {
            url += extension;
          }

          // absolute path
          url = path.resolve(dir, url);
          var basename = path.basename(url);
          var hasUnderscore = basename[0] === '_';

          try {
            fs.statSync(url);
            pathArray.push(url);
            continue;
          } catch (e) {}
          
          // without `_` prefix
          if (!hasUnderscore) {
            url = path.join(path.dirname(url), '_' + basename);

            try {
              fs.statSync(url);
              pathArray.push(url);
            } catch (e) {}
          }
        }

        return pathArray.map((url) => generateFakeModule( path.resolve(dir, url), id ));
      }

      // extract HTML/CSS files from ts modules
      // thus add watchers for their changes
      function extractPath(module) {
        var id = module.id;
        var dir = path.dirname(id);
        var source = module.originalCode;

        // thanks to https://github.com/cebor/rollup-plugin-angular
        var componentRegex = /@Component\(\s?{([\s\S]*)}\s?\)$/gm;
        var templateUrlRegex = /templateUrl\s*:\s*(.*)(,|\n)/g;
        var styleUrlsRegex = /styleUrls\s*:\s*\[([\s\S]*?)\]/g;
        var stringRegex = /(['"`])((?:[^\\]\\\1|.)*?)\1/g;
        var match = null;
        var pathArray = [];

        while ((match = componentRegex.exec(source))) {
          match[0].replace(templateUrlRegex, function (match, url) {
            pathArray.push(url.replace(/['"`,\s]/g, ''));
          }).replace(styleUrlsRegex, function (match, urls) {
            urls.trim().split(/\,\s*/).forEach(function (url) {
              url = url.trim();
              pathArray.push(url.replace(/['"`]/g, ''));
            });
          });
        }

        return pathArray.map((url) => generateFakeModule( path.resolve(dir, url), id ));
      }

      function generateFakeModule(id, relatedId) {
        var exists = true;

        try {
          fs.statSync(id);
        } catch (e) {
          throw (id + " does not exist!");
        }

        var fakeModule = htmlCSSCache.get(id);
        if (!fakeModule || !fakeModule.relatedIds) {
          fakeModule = {
            id: id,
            relatedIds: new Set([relatedId]),
            originalCode: fs.readFileSync(id, 'utf-8')
          };
        } else {
          fakeModule.relatedIds.add(relatedId);
        }

        // _global.css <== xx.component.ts <== xx.component.ts
        var relatedModule = htmlCSSCache.get(relatedId);
        var topRelatedIds = relatedModule && relatedModule.relatedIds || new Set();
        topRelatedIds.forEach(rid => fakeModule.relatedIds.add(rid));
        htmlCSSCache.set(id, fakeModule);

        if (STYLE_ECXT_REG.test(id)) {
          // watch related css files
          extractImportsFromCSS(fakeModule).map(addFileWatcher);
        }

        return fakeModule;
      }

      function addFileWatcher(module) {
        var id = module.id;

        if (!filewatchers.has(id)) {
          var watcher = new FileWatcher(id, module.originalCode, triggerRebuild, function () {
            filewatchers.delete(id);
          });

          if (watcher.fileExists) {
            filewatchers.set(id, watcher);
          }
        }
      }

      function build() {
        if (building) return;

        var start = Date.now();
        var initial = !watching;
        var opts = assign({}, options, cache ? { cache: cache } : {});

        emitter.emit('event', { code: 'BUILD_START' });

        building = true;

        return rollup.rollup(opts)
          .then(function (bundle) {
            // Save off bundle for re-use later
            cache = bundle;

            // we won't watch files from `node_modules`
            bundle.modules.filter(function (module) {
              var id = module.id;
              // skip plugin helper modules AND files from node_modules
              return module.id && module.id.indexOf('node_modules') < 0 && !/\0/.test(id);
            }).forEach(function (module) {
              extractPath(module).forEach(addFileWatcher);
              addFileWatcher(module);
            });

            if (options.targets) {
              return sequence(options.targets, function (target) {
                var mergedOptions = Object.assign({}, options, target);
                return bundle.write(mergedOptions);
              });
            }

            return bundle.write(options);
          })
          .then(function () {
            emitter.emit('event', {
              code: 'BUILD_END',
              duration: Date.now() - start,
              initial: initial
            });
          }, function (error) {
            emitter.emit('event', {
              code: 'ERROR',
              error: error
            });
          })
          .then(function () {
            building = false;
            if (rebuildScheduled) build();
          });
      }

      build();
    });

  return emitter;
}
